//========================================================================================
//  XInput.java
//    en:XInput Gamepad
//    ja:XInput Gamepad
//  Copyright (C) 2003-2023 Makoto Kamada
//
//  This file is part of the XEiJ (X68000 Emulator in Java).
//  You can use, modify and redistribute the XEiJ if the conditions are met.
//  Read the XEiJ License for more details.
//  https://stdkmd.net/xeij/
//========================================================================================

package xeij;

import java.util.*;

//class XInput
public final class XInput extends Thread {



  //動作制御

  protected Gamepad[] gamepads;  //Gamepadの配列
  protected ArrayList<GamepadListener> gamepadListeners;  //Gamepadリスナーのリスト
  protected boolean polling;  //true=ポーリング動作中

  //new XInput ()
  //  コンストラクタ
  public XInput () {
    //Gamepadの配列を作る
    gamepads = new Gamepad[4];
    //Gamepadを開く
    for (int index = 0; index < 4; index++) {
      gamepads[index] = new Gamepad (index);
    }
    //Gamepadリスナーのリストを作る
    gamepadListeners = new ArrayList<GamepadListener> ();
    //ポーリングを開始する
    polling = true;
    this.start ();
  }  //new XInput

  //end ()
  //  終了
  public void end () {
    //ポーリングを終了する
    if (polling) {
      polling = false;
      if (this.isAlive ()) {  //スレッドがある
        this.interrupt ();  //割り込む
        try {
          this.join ();  //止まるまで待つ
        } catch (InterruptedException ie) {
        }
      }
    }
    //Gamepadを閉じる
    for (int index = 0; index < 4; index++) {
      gamepads[index].close ();
    }
  }  //end

  //run ()
  //  ポーリング
  //  使用できるコントローラは1/100間隔、それ以外はおよそ1/3秒間隔でポーリングする
  @Override public void run () {
    while (polling) {
      boolean available = false;  //true=使用できるコントローラがある
      int counter = 0;  //0～31
      for (int index = 0; index < 4; index++) {
        Gamepad gamepad = gamepads[index];
        if (counter == 0 || gamepad.isAvailable ()) {
          available = available || gamepad.isAvailable ();
          if (gamepad.getState () &&  //状態を取得する。変化したかつ
              gamepadListeners.size () != 0) {  //リスナーがある
            if (gamepad.isConnected ()) {  //接続された
              for (GamepadListener listener : gamepadListeners) {
                listener.connected (gamepad);
              }
            } else if (gamepad.isDisconnected ()) {  //切断された
              for (GamepadListener listener : gamepadListeners) {
                listener.disconnected (gamepad);
              }
            }
            if (gamepad.isAvailable ()) {  //使用できる
              int buttonMasks = gamepad.getPressedButtonMasks ();
              if (buttonMasks != 0) {  //押されたボタンがある
                for (GamepadListener listener : gamepadListeners) {
                  listener.buttonPressed (gamepad, buttonMasks);
                }
              }
              buttonMasks = gamepad.getReleasedButtonMasks ();
              if (buttonMasks != 0) {  //離されたボタンがある
                for (GamepadListener listener : gamepadListeners) {
                  listener.buttonReleased (gamepad, buttonMasks);
                }
              }
              if (gamepad.isLeftStickMovedX ()) {  //左スティックがX方向に動いた
                for (GamepadListener listener : gamepadListeners) {
                  listener.leftStickMovedX (gamepad);
                }
              }
              if (gamepad.isLeftStickMovedY ()) {  //左スティックがY方向に動いた
                for (GamepadListener listener : gamepadListeners) {
                  listener.leftStickMovedY (gamepad);
                }
              }
              if (gamepad.isLeftTriggerMoved ()) {  //左トリガーが動いた
                for (GamepadListener listener : gamepadListeners) {
                  listener.leftTriggerMoved (gamepad);
                }
              }
              if (gamepad.isRightStickMovedX ()) {  //右スティックがX方向に動いた
                for (GamepadListener listener : gamepadListeners) {
                  listener.rightStickMovedX (gamepad);
                }
              }
              if (gamepad.isRightStickMovedY ()) {  //右スティックがY方向に動いた
                for (GamepadListener listener : gamepadListeners) {
                  listener.rightStickMovedY (gamepad);
                }
              }
              if (gamepad.isRightTriggerMoved ()) {  //右トリガーが動いた
                for (GamepadListener listener : gamepadListeners) {
                  listener.rightTriggerMoved (gamepad);
                }
              }
            }  //if 使用できる
          }  //if 変化したかつリスナーがある
        }  //if counter==0||isAvailable
      }  //for index
      try {
        if (available) {  //使用できるコントローラがある
          counter = (counter + 1) & 31;
          Thread.sleep (10L);
        } else {  //使用できるコントローラがない
          counter = 0;
          Thread.sleep (320L);
        }
      } catch (InterruptedException ie) {  //割り込まれた
        return;  //終了する
      }
    }  //while polling
  }  //run



  //状態取得

  //ボタン
  public static final int UP_BIT        = 0;  //上ボタン
  public static final int DOWN_BIT      = 1;  //下ボタン
  public static final int LEFT_BIT      = 2;  //左ボタン
  public static final int RIGHT_BIT     = 3;  //右ボタン
  public static final int START_BIT     = 4;  //STARTボタン
  public static final int BACK_BIT      = 5;  //BACKボタン
  public static final int LSTICK_BIT    = 6;  //左スティックボタン
  public static final int RSTICK_BIT    = 7;  //右スティックボタン
  public static final int LB_BIT        = 8;  //LBボタン
  public static final int RB_BIT        = 9;  //RBボタン
  public static final int A_BIT         = 12;  //Aボタン
  public static final int B_BIT         = 13;  //Bボタン
  public static final int X_BIT         = 14;  //Xボタン
  public static final int Y_BIT         = 15;  //Yボタン
  public static final int LSUP_BIT      = 16;  //左スティック上
  public static final int LSDOWN_BIT    = 17;  //左スティック下
  public static final int LSLEFT_BIT    = 18;  //左スティック左
  public static final int LSRIGHT_BIT   = 19;  //左スティック右
  public static final int RSUP_BIT      = 20;  //右スティック上
  public static final int RSDOWN_BIT    = 21;  //右スティック下
  public static final int RSLEFT_BIT    = 22;  //右スティック左
  public static final int RSRIGHT_BIT   = 23;  //右スティック右
  public static final int LTRIGGER_BIT  = 24;  //左トリガー
  public static final int RTRIGGER_BIT  = 25;  //右トリガー
  public static final int BUTTONS       = 26;  //ボタンの数

  public static final int UP_MASK       = 1 << UP_BIT;
  public static final int DOWN_MASK     = 1 << DOWN_BIT;
  public static final int LEFT_MASK     = 1 << LEFT_BIT;
  public static final int RIGHT_MASK    = 1 << RIGHT_BIT;
  public static final int START_MASK    = 1 << START_BIT;
  public static final int BACK_MASK     = 1 << BACK_BIT;
  public static final int LSTICK_MASK   = 1 << LSTICK_BIT;
  public static final int RSTICK_MASK   = 1 << RSTICK_BIT;
  public static final int LB_MASK       = 1 << LB_BIT;
  public static final int RB_MASK       = 1 << RB_BIT;
  public static final int A_MASK        = 1 << A_BIT;
  public static final int B_MASK        = 1 << B_BIT;
  public static final int X_MASK        = 1 << X_BIT;
  public static final int Y_MASK        = 1 << Y_BIT;
  public static final int LSUP_MASK     = 1 << LSUP_BIT;
  public static final int LSDOWN_MASK   = 1 << LSDOWN_BIT;
  public static final int LSLEFT_MASK   = 1 << LSLEFT_BIT;
  public static final int LSRIGHT_MASK  = 1 << LSRIGHT_BIT;
  public static final int RSUP_MASK     = 1 << RSUP_BIT;
  public static final int RSDOWN_MASK   = 1 << RSDOWN_BIT;
  public static final int RSLEFT_MASK   = 1 << RSLEFT_BIT;
  public static final int RSRIGHT_MASK  = 1 << RSRIGHT_BIT;
  public static final int LTRIGGER_MASK = 1 << LTRIGGER_BIT;
  public static final int RTRIGGER_MASK = 1 << RTRIGGER_BIT;

  public static final String BIT_TO_TEXT[] = new String[] {
    "UP", "DOWN", "LEFT", "RIGHT", "START", "BACK", "LSTICK", "RSTICK",
    "LB", "RB", "(10)", "(11)", "A", "B", "X", "Y",
    "LSUP", "LSDOWN", "LSLEFT", "LSRIGHT", "RSUP", "RSDOWN", "RSLEFT", "RSRIGHT",
    "LTRIGGER", "RTRIGGER",
  };

  //buttonMasks = getButtonMasks (index)
  //  すべての押されているボタンのマスクを返す
  //  index  コントローラのインデックス
  //  buttonMasks  すべての押されているボタンのマスク
  public int getButtonMasks (int index) {
    return gamepads[index].getButtonMasks ();
  }

  //leftStickX = getLeftStickX (index)
  //  左スティックのX方向の位置を返す
  //  index  コントローラのインデックス
  //  leftStickX  左スティックのX方向の位置
  public int getLeftStickX (int index) {
    return gamepads[index].getLeftStickX ();
  }

  //leftStickY = getLeftStickY (index)
  //  左スティックのY方向の位置を返す
  //  index  コントローラのインデックス
  //  leftStickY  左スティックのY方向の位置
  public int getLeftStickY (int index) {
    return gamepads[index].getLeftStickY ();
  }

  //leftTrigger = getLeftTrigger (index)
  //  左トリガーの位置を返す
  //  index  コントローラのインデックス
  //  leftTrigger  左トリガーの位置
  public int getLeftTrigger (int index) {
    return gamepads[index].getLeftTrigger ();
  }

  //rightStickX = getRightStickX (index)
  //  右スティックのX方向の位置を返す
  //  index  コントローラのインデックス
  //  rightStickX  右スティックのX方向の位置
  public int getRightStickX (int index) {
    return gamepads[index].getRightStickX ();
  }

  //rightStickY = getRightStickY (index)
  //  右スティックのY方向の位置を返す
  //  index  コントローラのインデックス
  //  rightStickY  右スティックのY方向の位置
  public int getRightStickY (int index) {
    return gamepads[index].getRightStickY ();
  }

  //rightTrigger = getRightTrigger (index)
  //  右トリガーの位置を返す
  //  index  コントローラのインデックス
  //  rightTrigger  右トリガーの位置
  public int getRightTrigger (int index) {
    return gamepads[index].getRightTrigger ();
  }

  //available = isAvailable (index)
  //  使用できるかを返す
  //  index  コントローラのインデックス
  //  available  使用できるか
  public boolean isAvailable (int index) {
    return gamepads[index].isAvailable ();
  }

/*
  //setCenterOfLeftStick (index, centerOfLeftStickX, centerOfLeftStickY)
  //  左スティックの中央の位置を設定する
  //  index  コントローラのインデックス
  //  centerOfLeftStickX  左スティックのX方向の中央の位置
  //  centerOfLeftStickY  左スティックのY方向の中央の位置
  public void setCenterOfLeftStick (int index, int centerOfLeftStickX, int centerOfLeftStickY) {
    gamepads[index].setCenterOfLeftStick (centerOfLeftStickX, centerOfLeftStickY);
  }

  //setCenterOfRightStick (index, centerOfRightStickX, centerOfRightStickY)
  //  右スティックの中央の位置を設定する
  //  index  コントローラのインデックス
  //  centerOfRightStickX  右スティックのX方向の中央の位置
  //  centerOfRightStickY  右スティックのY方向の中央の位置
  public void setCenterOfRightStick (int index, int centerOfRightStickX, int centerOfRightStickY) {
    gamepads[index].setCenterOfRightStick (centerOfRightStickX, centerOfRightStickY);
  }

  //setPlayOfLeftStick (index, playOfLeftStickX, playOfLeftStickY)
  //  左スティックの遊びを設定する
  //  index  コントローラのインデックス
  //  playOfLeftStickX  左スティックのX方向の遊び
  //  playOfLeftStickY  左スティックのY方向の遊び
  public void setPlayOfLeftStick (int index, int playOfLeftStickX, int playOfLeftStickY) {
    gamepads[index].setPlayOfLeftStick (playOfLeftStickX, playOfLeftStickY);
  }

  //setPlayOfRightStick (index, playOfRightStickX, playOfRightStickY)
  //  右スティックの遊びを設定する
  //  index  コントローラのインデックス
  //  playOfRightStickX  右スティックのX方向の遊び
  //  playOfRightStickY  右スティックのY方向の遊び
  public void setPlayOfRightStick (int index, int playOfRightStickX, int playOfRightStickY) {
    gamepads[index].setPlayOfRightStick (playOfRightStickX, playOfRightStickY);
  }
*/

  //setThresholdOfLeftStick (index, thresholdOfLeftStick)
  //  左スティックの閾値を設定する
  //  index  コントローラのインデックス
  //  thresholdOfLeftStick  左スティックの閾値
  public void setThresholdOfLeftStick (int index, int thresholdOfLeftStick) {
    gamepads[index].setThresholdOfLeftStick (thresholdOfLeftStick);
  }

  //setThresholdOfRightStick (index, thresholdOfRightStick)
  //  右スティックの閾値を設定する
  //  index  コントローラのインデックス
  //  thresholdOfRightStick  右スティックの閾値
  public void setThresholdOfRightStick (int index, int thresholdOfRightStick) {
    gamepads[index].setThresholdOfRightStick (thresholdOfRightStick);
  }

  //setThresholdOfLeftTrigger (index, thresholdOfLeftTrigger)
  //  左トリガーの閾値を設定する
  //  index  コントローラのインデックス
  //  thresholdOfLeftTrigger  左トリガーの閾値
  public void setThresholdOfLeftTrigger (int index, int thresholdOfLeftTrigger) {
    gamepads[index].setThresholdOfLeftTrigger (thresholdOfLeftTrigger);
  }

  //setThresholdOfRightTrigger (index, thresholdOfRightTrigger)
  //  右トリガーの閾値を設定する
  //  index  コントローラのインデックス
  //  thresholdOfRightTrigger  右トリガーの閾値
  public void setThresholdOfRightTrigger (int index, int thresholdOfRightTrigger) {
    gamepads[index].setThresholdOfRightTrigger (thresholdOfRightTrigger);
  }



  //Gamepadリスナー

  //interface XInput.GamepadListener
  public interface GamepadListener {
    public void connected (Gamepad gamepad);  //接続された
    public void disconnected (Gamepad gamepad);  //切断された
    public void buttonPressed (Gamepad gamepad, int buttonMasks);  //ボタンが押された
    public void buttonReleased (Gamepad gamepad, int buttonMasks);  //ボタンが離された
    public void leftStickMovedX (Gamepad gamepad);  //左スティックがX方向に動いた
    public void leftStickMovedY (Gamepad gamepad);  //左スティックがY方向に動いた
    public void leftTriggerMoved (Gamepad gamepad);  //左トリガーが動いた
    public void rightStickMovedX (Gamepad gamepad);  //右スティックがX方向に動いた
    public void rightStickMovedY (Gamepad gamepad);  //右スティックがY方向に動いた
    public void rightTriggerMoved (Gamepad gamepad);  //右トリガーが動いた
  }

  //class XInput.GamepadAdapter
  public static class GamepadAdapter implements GamepadListener {
    @Override public void connected (Gamepad gamepad) {
    }
    @Override public void disconnected (Gamepad gamepad) {
    }
    @Override public void buttonPressed (Gamepad gamepad, int buttonMasks) {
    }
    @Override public void buttonReleased (Gamepad gamepad, int buttonMasks) {
    }
    @Override public void leftStickMovedX (Gamepad gamepad) {
    }
    @Override public void leftStickMovedY (Gamepad gamepad) {
    }
    @Override public void leftTriggerMoved (Gamepad gamepad) {
    }
    @Override public void rightStickMovedX (Gamepad gamepad) {
    }
    @Override public void rightStickMovedY (Gamepad gamepad) {
    }
    @Override public void rightTriggerMoved (Gamepad gamepad) {
    }
  }

  //addGamepadListener (listener)
  //  Gamepadリスナーを追加する
  //  listener  追加するGamepadリスナー
  public void addGamepadListener (GamepadListener listener) {
    if (listener != null && !gamepadListeners.contains (listener)) {
      gamepadListeners.add (listener);
    }
  }

  //removeGamepadListener (listener)
  //  Gamepadリスナーを削除する
  //  listener  削除するGamepadリスナー
  public void removeGamepadListener (GamepadListener listener) {
    gamepadListeners.remove (listener);
  }

  //removeGamepadListeners ()
  //  すべてのGamepadリスナーを削除する
  public void removeGamepadListeners () {
    gamepadListeners.clear ();
  }

  //getGamepadListeners ()
  //  すべてのGamepadリスナーを取得する
  public GamepadListener[] getGamepadListeners () {
    return gamepadListeners.toArray (new GamepadListener[gamepadListeners.size ()]);
  }



  //class XInput.Gamepad
  public static class Gamepad implements AutoCloseable {

    private long xiwork;  //ワークエリアを指すポインタ

    //new Gamepad (index)
    //  コンストラクタ
    //  index  コントローラのインデックス。0～3
    public Gamepad (int index) {
      open (index);
    }

    //close ()
    //  閉じる
    @Override public native void close ();

    //buttonMasks = getButtonMasks ()
    //  getState()で取得したすべての押されているボタンのマスクを返す
    //  buttonMasks  すべての押されているボタンのマスク
    public native int getButtonMasks ();

    //index = getIndex ()
    //  コントローラのインデックスを返す
    //  index  コントローラのインデックス
    public native int getIndex ();

    //leftStickX = getLeftStickX ()
    //  getState()で取得した左スティックのX方向の位置を返す
    //  leftStickX  左スティックのX方向の位置
    public native int getLeftStickX ();

    //leftStickY = getLeftStickY ()
    //  getState()で取得した左スティックのY方向の位置を返す
    //  leftStickY  左スティックのY方向の位置
    public native int getLeftStickY ();

    //leftTrigger = getLeftTrigger ()
    //  getState()で取得した左トリガーの位置を返す
    //  leftTrigger  左トリガーの位置
    public native int getLeftTrigger ();

    //pressedButtonMasks = getPressedButtonMasks ()
    //  getState()で取得したすべての押されたボタンのマスクを返す
    //  pressedButtonMasks  すべての押されたボタンのマスク
    public native int getPressedButtonMasks ();

    //releasedButtonMasks = getReleasedButtonMasks ()
    //  getState()で取得したすべての離されたボタンのマスクを返す
    //  releasedButtonMasks  すべての離されたボタンのマスク
    public native int getReleasedButtonMasks ();

    //rightStickX = getRightStickX ()
    //  getState()で取得した右スティックのX方向の位置を返す
    //  rightStickX  右スティックのX方向の位置
    public native int getRightStickX ();

    //rightStickY = getRightStickY ()
    //  getState()で取得した右スティックのY方向の位置を返す
    //  rightStickY  右スティックのY方向の位置
    public native int getRightStickY ();

    //rightTrigger = getRightTrigger ()
    //  getState()で取得した右トリガーの位置を返す
    //  rightTrigger  右トリガーの位置
    public native int getRightTrigger ();

    //changed = getState ()
    //  コントローラの状態を取得する
    //  changed  false=変化しなかった,true=変化した。使用可能とは限らない
    public native boolean getState ();

    //available = isAvailable ()
    //  使用できるかを返す
    //  available  使用できるか
    public native boolean isAvailable ();

    //connected = isConnected ()
    //  接続されたかを返す
    //  connected  接続されたか
    public native boolean isConnected ();

    //disconnected = isDisconnected ()
    //  切断されたかを返す
    //  disconnected  切断されたか
    public native boolean isDisconnected ();

    //leftStickMovedX = isLeftStickMovedX ()
    //  getState()で取得した左スティックがX方向に動いたかを返す
    //  leftStickMovedX  左スティックがX方向に動いたか
    public native boolean isLeftStickMovedX ();

    //leftStickMovedY = isLeftStickMovedY ()
    //  getState()で取得した左スティックがY方向に動いたかを返す
    //  leftStickMovedY  左スティックがY方向に動いたか
    public native boolean isLeftStickMovedY ();

    //leftTriggerMoved = isLeftTriggerMoved ()
    //  getState()で取得した左トリガーが動いたかを返す
    //  leftTriggerMoved  左トリガーが動いたか
    public native boolean isLeftTriggerMoved ();

    //rightStickMovedX = isRightStickMovedX ()
    //  getState()で取得した右スティックがX方向に動いたかを返す
    //  rightStickMovedX  右スティックがX方向に動いたか
    public native boolean isRightStickMovedX ();

    //rightStickMovedY = isRightStickMovedY ()
    //  getState()で取得した右スティックがY方向に動いたかを返す
    //  rightStickMovedY  右スティックがY方向に動いたか
    public native boolean isRightStickMovedY ();

    //rightTriggerMoved = isRightTriggerMoved ()
    //  getState()で取得した右トリガーが動いたかを返す
    //  rightTriggerMoved  右トリガーが動いたか
    public native boolean isRightTriggerMoved ();

    //open (index)
    //  開く
    //  index  コントローラのインデックス。0～3
    private native void open (int index);

/*
    //setCenterOfLeftStick (centerOfLeftStickX, centerOfLeftStickY)
    //  左スティックの中央の位置を設定する
    //  centerOfLeftStickX  左スティックのX方向の中央の位置
    //  centerOfLeftStickY  左スティックのY方向の中央の位置
    public native void setCenterOfLeftStick (int centerOfLeftStickX, int centerOfLeftStickY);

    //setCenterOfRightStick (centerOfRightStickX, centerOfRightStickY)
    //  右スティックの中央の位置を設定する
    //  centerOfRightStickX  右スティックのX方向の中央の位置
    //  centerOfRightStickY  右スティックのY方向の中央の位置
    public native void setCenterOfRightStick (int centerOfRightStickX, int centerOfRightStickY);

    //setPlayOfLeftStick (playOfLeftStickX, playOfLeftStickY)
    //  左スティックの遊びを設定する
    //  playOfLeftStickX  左スティックのX方向の遊び
    //  playOfLeftStickY  左スティックのY方向の遊び
    public native void setPlayOfLeftStick (int playOfLeftStickX, int playOfLeftStickY);

    //setPlayOfRightStick (playOfRightStickX, playOfRightStickY)
    //  右スティックの遊びを設定する
    //  playOfRightStickX  右スティックのX方向の遊び
    //  playOfRightStickY  右スティックのY方向の遊び
    public native void setPlayOfRightStick (int playOfRightStickX, int playOfRightStickY);
*/

    //------------------------------------------------------------------------
    //setThresholdOfLeftStick (thresholdOfLeftStick)
    //  左スティックの閾値を設定する
    //  thresholdOfLeftStick  左スティックの閾値
    public native void setThresholdOfLeftStick (int thresholdOfLeftStick);

    //------------------------------------------------------------------------
    //setThresholdOfRightStick (thresholdOfRightStick)
    //  右スティックの閾値を設定する
    //  thresholdOfRightStick  右スティックの閾値
    public native void setThresholdOfRightStick (int thresholdOfRightStick);

    //------------------------------------------------------------------------
    //setThresholdOfLeftTrigger (thresholdOfLeftTrigger)
    //  左トリガーの閾値を設定する
    //  thresholdOfLeftTrigger  左トリガーの閾値
    public native void setThresholdOfLeftTrigger (int thresholdOfLeftTrigger);

    //------------------------------------------------------------------------
    //setThresholdOfRightTrigger (thresholdOfRightTrigger)
    //  右トリガーの閾値を設定する
    //  thresholdOfRightTrigger  右トリガーの閾値
    public native void setThresholdOfRightTrigger (int thresholdOfRightTrigger);

  }  //class Gamepad



}  //class XInput
